/*
  main.c for LCDd

  Contains main(), plus signal callback functions and a help screen.

  Program init, command-line handling, and the main loop are
  implemented here.  Also, minimal data about the program such as
  the revision number.

  Some of this stuff should probably be move elsewhere eventually,
  such as command-line handling and the main loop.  main() is supposed
  to be "dumb".
  
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "../shared/debug.h"

#include "drivers/lcd.h"
#include "sock.h"
#include "clients.h"
#include "screenlist.h"
#include "screen.h"
#include "parse.h"
#include "render.h"
#include "serverscreens.h"
#include "input.h"
#include "main.h"


char *version = VERSION;
char *build_date = __DATE__;


void exit_program(int val);
void HelpScreen();


int main(int argc, char **argv)
{
   // TODO:  Use a config file!
   //char cfgfile[256] = "/etc/LCDd.cf";
   int i, err;
   int daemon_mode=1;
   int child;
   screen *s = NULL;
   char *str, *ing;  // strings for commandline handling
   

   
   // Ctrl-C will cause a clean exit...
   signal(SIGINT, exit_program);
   // and "kill"...
   signal(SIGTERM, exit_program);
   // and "kill -HUP" (hangup)...
   signal(SIGHUP, exit_program);
   // and just in case, "kill -KILL" (which cannot be trapped; but oh well)
   signal(SIGKILL, exit_program);


   lcd_init("");

   
   // Parse the command line now...
   // TODO:  Move this to a separate function?
   for(i=1; i<argc; i++)
   {
      if(0 == strcmp(argv[i], "-d") ||
	 0 == strcmp(argv[i], "--driver"))
      {
	 if(argc <= i+1) HelpScreen();
	 str = argv[++i];
	 ing = NULL;
	 if(argc > i+1)
	 {
	    if(strcmp(argv[i+1], "-d") &&
	       strcmp(argv[i+1], "--driver"))
	    //if(argv[i+1][0] != '.')
	    {
	       ing = argv[++i];
	    }
	    //else i++;
	 }
	 err = lcd_add_driver(str, ing);
	 if(err <= 0)
	 {
	    printf("Error loading driver %s.  Continuing anyway...\n",
		   str);
	 }
	 if((err > 0) && (0 == strcmp(str, "curses"))) daemon_mode=0;
      }
      else
      {
	 // otherwise...  Get help!
	 printf("Invalid parameter: %s\n", argv[i]);
	 HelpScreen();
      }
   }

   

   // Now init a bunch of required stuff...

   if(sock_create_server() <= 0)
   {
      printf("Error opening socket.\n");
      return 1;
   }

   if(client_init() < 0)
   {
      printf("Error initializing client list\n");
      return 1;
   }
   
   if(screenlist_init() < 0)
   {
      printf("Error initializing screen list\n");
      return 1;
   }

   // Make sure the server screen shows up every once in a while..
   if(server_screen_init() < 0)
   {
      printf("Error initializing server screens\n");
      return 1;
   }



   
   // Now, go into daemon mode...
   #ifndef DEBUG
   if(daemon_mode)
   {
     if ((child = fork()) != 0) exit(0);             /* PARENT EXITS */
     // This line removed because it eats error messages...
     //setsid();                                       /* RELEASE TTY */
   }
   #endif



   
   // Main loop...
   while(1)
   {
      sock_poll_clients();
      parse_all_client_messages();
      handle_input();

      // TODO:  Move this code to screenlist.c...
      // ... it should just say "handle_screens();"
      // Timer gets reset by screenlist_next()
      timer++;
      // this line's here because s was getting overwritten at one time...
      //s = screenlist_current();
      if(s  &&  (timer >= s->duration))
      {
	 screenlist_next();
      }
      // Just in case it gets out of hand...
      if(timer >= 0x10000) timer = 0;
      update_server_screen(timer);
      s = screenlist_current();
	 
      // render something here...
      if(s) draw_screen(s, timer);
      else  no_screen_screen(timer);
      
      usleep(TIME_UNIT);
   }


   // Quit!
   exit_program(0);
   
   return 0;
}


void exit_program(int val)
{
   // Say goodbye!
   goodbye_screen();
   
   // Can go anywhere...
   lcd_shutdown();

   // TODO: These things shouldn't be so interdependent.  The order
   // things are shut down in shouldn't matter...
   
   // Must come before screenlist_shutdown
   client_shutdown();
   // Must come after client_shutdown
   screenlist_shutdown();
   // Should come after client_shutdown
   sock_close_all();
   
   exit(0);

}


void HelpScreen()
{
  printf("LCDproc server daemon, %s\n", version);
  printf("Usage: lcdproc [-d driver [args]]\n");
  printf("\tOptions in []'s are optional.\n");
  printf("\t-d driver is the output driver to use:\n");
  printf("\t\tMtxOrb, curses, text, HD44780, joy\n");
  printf("\targs will be passed to the driver for init\n");
  printf("\n");
  printf("\tUse \"man LCDd\" for more info.\n");
  printf("Example:\n");
  printf("\tLCDd -d MtxOrb \"dev /dev/lcd\" -d joy\n");
  printf("\n");
  exit(0);
}

